@@ -11,55 +11,37 @@
  * 26 Jun 96 - Version 3.0 by Eric S. Raymond (Full GIF89 support)
  *****************************************************************************/
 
-#ifdef HAVE_CONFIG_H
-#include <config.h>
-#endif
+#include <unistd.h>
+#include <stdint.h>
+#include <stdlib.h>
+#include <stdio.h>
+#include <string.h>
+#include <fcntl.h>
 
-#ifdef __MSDOS__
+#if (defined (__MSDOS__) || defined(WINDOWS32))  && !defined(__DJGPP__) && !defined(__GNUC__)
 #include <io.h>
-#include <alloc.h>
 #include <sys\stat.h>
 #else
 #include <sys/types.h>
 #include <sys/stat.h>
-#ifdef R6000
-
-/* FIXME: What is sys/mode.h?  Can we substitute a check for this file rather
- * than a check based on machine type?
- */
-#include <sys/mode.h>
-#endif
 #endif /* __MSDOS__ */
 
-#ifdef HAVE_IO_H
-#include <io.h>
-#endif
-
-#ifdef HAVE_FCNTL_H
-#include <fcntl.h>
-#endif /* HAVE_FCNTL_H */
-#ifdef HAVE_UNISTD_H
-#include <unistd.h>
-#endif /* HAVE_UNISTD_H */
-#include <stdlib.h>
-#include <stdio.h>
-#include <string.h>
 #include "gif_lib.h"
 #include "gif_lib_private.h"
 
-/* #define DEBUG_NO_PREFIX                  Dump only compressed data. */
-
 /* Masks given codes to BitsPerPixel, to make sure all codes are in range: */
-static GifPixelType CodeMask[] = {
+/*@+charint@*/
+static const GifPixelType CodeMask[] = {
     0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f, 0xff
 };
+/*@-charint@*/
 
-static char GifVersionPrefix[GIF_STAMP_LEN + 1] = GIF87_STAMP;
+static char GifVersionPrefix[GIF_STAMP_LEN + 1];
 
 #define WRITE(_gif,_buf,_len)   \
   (((GifFilePrivateType*)_gif->Private)->Write ?    \
    ((GifFilePrivateType*)_gif->Private)->Write(_gif,_buf,_len) :    \
-   fwrite(_buf, 1, _len, ((GifFilePrivateType*)_gif->Private)->File))
+   fwrite(_buf, 1, (size_t)_len, ((GifFilePrivateType*)_gif->Private)->File))
 
 static int EGifPutWord(int Word, GifFileType * GifFile);
 static int EGifSetupCompress(GifFileType * GifFile);
@@ -77,23 +59,17 @@ static int EGifBufferedOutput(GifFileType * GifFile, GifByteType * Buf,
  *****************************************************************************/
 GifFileType *
 EGifOpenFileName(const char *FileName,
-                 int TestExistance) {
+                 bool TestExistance) {
 
     int FileHandle;
     GifFileType *GifFile;
 
     if (TestExistance)
-        FileHandle = open(FileName, O_WRONLY | O_CREAT | O_EXCL
-#ifdef __MSDOS__
-                          | O_BINARY
-#endif /* __MSDOS__ */
-                          , S_IREAD | S_IWRITE);
+        FileHandle = open(FileName, O_WRONLY | O_CREAT | O_EXCL, 
+			  S_IREAD | S_IWRITE);
     else
-        FileHandle = open(FileName, O_WRONLY | O_CREAT | O_TRUNC
-#ifdef __MSDOS__
-                          | O_BINARY
-#endif /* __MSDOS__ */
-                          , S_IREAD | S_IWRITE);
+        FileHandle = open(FileName, O_WRONLY | O_CREAT | O_TRUNC, 
+			  S_IREAD | S_IWRITE);
 
     if (FileHandle == -1) {
         _GifError = E_GIF_ERR_OPEN_FAILED;
@@ -139,24 +115,23 @@ EGifOpenFileHandle(int FileHandle) {
         return NULL;
     }
 
-#ifdef __MSDOS__
+#if defined(__MSDOS__) || defined(WINDOWS32) || defined(_OPEN_BINARY)
     setmode(FileHandle, O_BINARY);    /* Make sure it is in binary mode. */
 #endif /* __MSDOS__ */
 
     f = fdopen(FileHandle, "wb");    /* Make it into a stream: */
 
-#ifdef __MSDOS__
+    /* MSDOS and Windows32 requires this, no reason not to do it under Unix */
     setvbuf(f, NULL, _IOFBF, GIF_FILE_BUFFER_SIZE);    /* And inc. stream
                                                         * buffer. */
-#endif /* __MSDOS__ */
 
-    GifFile->Private = (VoidPtr)Private;
+    GifFile->Private = (void *)Private;
     Private->FileHandle = FileHandle;
     Private->File = f;
     Private->FileState = FILE_STATE_WRITE;
 
     Private->Write = (OutputFunc) 0;    /* No user write routine (MRB) */
-    GifFile->UserData = (VoidPtr) 0;    /* No user write handle (MRB) */
+    GifFile->UserData = (void *)NULL;    /* No user write handle (MRB) */
 
     _GifError = 0;
 
@@ -189,7 +164,15 @@ EGifOpen(void *userData,
         return NULL;
     }
 
-    GifFile->Private = (VoidPtr) Private;
+    Private->HashTable = _InitHashTable();
+    if (Private->HashTable == NULL) {
+        free (GifFile);
+        free (Private);
+        _GifError = E_GIF_ERR_NOT_ENOUGH_MEM;
+        return NULL;
+    }
+
+    GifFile->Private = (void *)Private;
     Private->FileHandle = 0;
     Private->File = (FILE *) 0;
     Private->FileState = FILE_STATE_WRITE;
@@ -224,9 +207,10 @@ EGifPutScreenDesc(GifFileType * GifFile,
                   int BackGround,
                   const ColorMapObject * ColorMap) {
 
-    int i;
     GifByteType Buf[3];
     GifFilePrivateType *Private = (GifFilePrivateType *) GifFile->Private;
+    char *write_version;
+    int i, j;
 
     if (Private->FileState & FILE_STATE_SCREEN) {
         /* If already has screen descriptor - something is wrong! */
@@ -239,14 +223,42 @@ EGifPutScreenDesc(GifFileType * GifFile,
         return GIF_ERROR;
     }
 
-/* First write the version prefix into the file. */
-#ifndef DEBUG_NO_PREFIX
-    if (WRITE(GifFile, (unsigned char *)GifVersionPrefix,
-              strlen(GifVersionPrefix)) != strlen(GifVersionPrefix)) {
+    /* Bulletproofing - always write GIF89 if we need to */
+    for (i = 0; i < GifFile->ImageCount; i++) {
+        for (j = 0; j < GifFile->SavedImages[i].ExtensionBlockCount; j++) {
+            int function =
+               GifFile->SavedImages[i].ExtensionBlocks[j].Function;
+
+            if (function == COMMENT_EXT_FUNC_CODE
+                || function == GRAPHICS_EXT_FUNC_CODE
+                || function == PLAINTEXT_EXT_FUNC_CODE
+                || function == APPLICATION_EXT_FUNC_CODE)
+                Private->gif89 = true;
+        }
+    }
+
+    /*
+     * Older versions of the library didn't compute which version 
+     * the image's extension blocks require here, but rather in EGifSpew(),
+     * and it was done in a non-thread-safe way. The default for sequential 
+     * writes was GIF87 but you could override it with EGifSetGifVersion.
+     * This new code lets you shoot yourself in the foot if you really
+     * want to, but defaults to the oldest version that will carry the
+     * extensions.
+     */
+    if (GifVersionPrefix[0] != '\0')
+	write_version = GifVersionPrefix;
+    else if (Private->gif89)
+	write_version = GIF89_STAMP;
+    else
+	write_version = GIF87_STAMP;
+
+    /* First write the version prefix into the file. */
+    if (WRITE(GifFile, (unsigned char *)write_version,
+              strlen(write_version)) != strlen(write_version)) {
         _GifError = E_GIF_ERR_WRITE_FAILED;
         return GIF_ERROR;
     }
-#endif /* DEBUG_NO_PREFIX */
 
     GifFile->SWidth = Width;
     GifFile->SHeight = Height;
@@ -280,13 +292,11 @@ EGifPutScreenDesc(GifFileType * GifFile,
                                                             color table. */
     Buf[1] = BackGround;    /* Index into the ColorTable for background color */
     Buf[2] = 0;             /* Pixel Aspect Ratio */
-#ifndef DEBUG_NO_PREFIX
     WRITE(GifFile, Buf, 3);
-#endif /* DEBUG_NO_PREFIX */
 
     /* If we have Global color map - dump it also: */
-#ifndef DEBUG_NO_PREFIX
-    if (ColorMap != NULL)
+    if (ColorMap != NULL) {
+	int i;
         for (i = 0; i < ColorMap->ColorCount; i++) {
             /* Put the ColorMap out also: */
             Buf[0] = ColorMap->Colors[i].Red;
@@ -297,7 +307,7 @@ EGifPutScreenDesc(GifFileType * GifFile,
                 return GIF_ERROR;
             }
         }
-#endif /* DEBUG_NO_PREFIX */
+    }
 
     /* Mark this file as has screen descriptor, and no pixel written yet: */
     Private->FileState |= FILE_STATE_SCREEN;
@@ -315,15 +325,14 @@ EGifPutImageDesc(GifFileType * GifFile,
                  int Top,
                  int Width,
                  int Height,
-                 int Interlace,
+                 bool Interlace,
                  const ColorMapObject * ColorMap) {
 
-    int i;
     GifByteType Buf[3];
     GifFilePrivateType *Private = (GifFilePrivateType *)GifFile->Private;
 
     if (Private->FileState & FILE_STATE_IMAGE &&
-#if defined(__MSDOS__) || defined(__GNUC__)
+#if defined(__MSDOS__) || defined(WINDOWS32) || defined(__GNUC__)
         Private->PixelCount > 0xffff0000UL) {
 #else
         Private->PixelCount > 0xffff0000) {
@@ -355,9 +364,7 @@ EGifPutImageDesc(GifFileType * GifFile,
 
     /* Put the image descriptor into the file: */
     Buf[0] = ',';    /* Image seperator character. */
-#ifndef DEBUG_NO_PREFIX
     WRITE(GifFile, Buf, 1);
-#endif /* DEBUG_NO_PREFIX */
     EGifPutWord(Left, GifFile);
     EGifPutWord(Top, GifFile);
     EGifPutWord(Width, GifFile);
@@ -365,13 +372,11 @@ EGifPutImageDesc(GifFileType * GifFile,
     Buf[0] = (ColorMap ? 0x80 : 0x00) |
        (Interlace ? 0x40 : 0x00) |
        (ColorMap ? ColorMap->BitsPerPixel - 1 : 0);
-#ifndef DEBUG_NO_PREFIX
     WRITE(GifFile, Buf, 1);
-#endif /* DEBUG_NO_PREFIX */
 
     /* If we have Global color map - dump it also: */
-#ifndef DEBUG_NO_PREFIX
-    if (ColorMap != NULL)
+    if (ColorMap != NULL) {
+	int i;
         for (i = 0; i < ColorMap->ColorCount; i++) {
             /* Put the ColorMap out also: */
             Buf[0] = ColorMap->Colors[i].Red;
@@ -382,7 +387,7 @@ EGifPutImageDesc(GifFileType * GifFile,
                 return GIF_ERROR;
             }
         }
-#endif /* DEBUG_NO_PREFIX */
+    }
     if (GifFile->SColorMap == NULL && GifFile->Image.ColorMap == NULL) {
         _GifError = E_GIF_ERR_NO_COLOR_MAP;
         return GIF_ERROR;
@@ -514,7 +519,7 @@ int
 EGifPutExtensionFirst(GifFileType * GifFile,
                       int ExtCode,
                       int ExtLen,
-                      const VoidPtr Extension) {
+                      const void *Extension) {
 
     GifByteType Buf[3];
     GifFilePrivateType *Private = (GifFilePrivateType *)GifFile->Private;
@@ -546,7 +551,7 @@ int
 EGifPutExtensionNext(GifFileType * GifFile,
                      int ExtCode,
                      int ExtLen,
-                     const VoidPtr Extension) {
+                     const void *Extension) {
 
     GifByteType Buf;
     GifFilePrivateType *Private = (GifFilePrivateType *)GifFile->Private;
@@ -571,7 +576,7 @@ int
 EGifPutExtensionLast(GifFileType * GifFile,
                      int ExtCode,
                      int ExtLen,
-                     const VoidPtr Extension) {
+                     const void *Extension) {
 
     GifByteType Buf;
     GifFilePrivateType *Private = (GifFilePrivateType *)GifFile->Private;
@@ -606,7 +611,7 @@ int
 EGifPutExtension(GifFileType * GifFile,
                  int ExtCode,
                  int ExtLen,
-                 const VoidPtr Extension) {
+                 const void *Extension) {
 
     GifByteType Buf[3];
     GifFilePrivateType *Private = (GifFilePrivateType *)GifFile->Private;
@@ -708,6 +713,8 @@ EGifCloseFile(GifFileType * GifFile) {
         return GIF_ERROR;
 
     Private = (GifFilePrivateType *) GifFile->Private;
+    if (Private == NULL)
+	return GIF_ERROR;
     if (!IS_WRITEABLE(Private)) {
         /* This file was NOT open for writing: */
         _GifError = E_GIF_ERR_NOT_WRITEABLE;
@@ -719,10 +726,14 @@ EGifCloseFile(GifFileType * GifFile) {
     Buf = ';';
     WRITE(GifFile, &Buf, 1);
 
-    if (GifFile->Image.ColorMap)
+    if (GifFile->Image.ColorMap) {
         FreeMapObject(GifFile->Image.ColorMap);
-    if (GifFile->SColorMap)
+        GifFile->Image.ColorMap = NULL;
+    }
+    if (GifFile->SColorMap) {
         FreeMapObject(GifFile->SColorMap);
+        GifFile->SColorMap = NULL;
+    }
     if (Private) {
         if (Private->HashTable) {
             free((char *) Private->HashTable);
@@ -749,14 +760,10 @@ EGifPutWord(int Word,
 
     c[0] = Word & 0xff;
     c[1] = (Word >> 8) & 0xff;
-#ifndef DEBUG_NO_PREFIX
     if (WRITE(GifFile, c, 2) == 2)
         return GIF_OK;
     else
         return GIF_ERROR;
-#else
-    return GIF_OK;
-#endif /* DEBUG_NO_PREFIX */
 }
 
 /******************************************************************************
@@ -832,7 +839,7 @@ EGifCompressLine(GifFileType * GifFile,
         /* Form a new unique key to search hash table for the code combines 
          * CrntCode as Prefix string with Pixel as postfix char.
          */
-        NewKey = (((unsigned long) CrntCode) << 8) + Pixel;
+        NewKey = (((uint32_t) CrntCode) << 8) + Pixel;
         if ((NewCode = _ExistsHashTable(HashTable, NewKey)) >= 0) {
             /* This Key is already there, or the string is old one, so
              * simple take new code as our CrntCode:
@@ -986,39 +993,17 @@ EGifBufferedOutput(GifFileType * GifFile,
 int
 EGifSpew(GifFileType * GifFileOut) {
 
-    int i, j, gif89 = FALSE;
+    int i, j; 
     int bOff;   /* Block Offset for adding sub blocks in Extensions */
-    char SavedStamp[GIF_STAMP_LEN + 1];
 
-    for (i = 0; i < GifFileOut->ImageCount; i++) {
-        for (j = 0; j < GifFileOut->SavedImages[i].ExtensionBlockCount; j++) {
-            int function =
-               GifFileOut->SavedImages[i].ExtensionBlocks[j].Function;
-
-            if (function == COMMENT_EXT_FUNC_CODE
-                || function == GRAPHICS_EXT_FUNC_CODE
-                || function == PLAINTEXT_EXT_FUNC_CODE
-                || function == APPLICATION_EXT_FUNC_CODE)
-                gif89 = TRUE;
-        }
-    }
-
-    strncpy(SavedStamp, GifVersionPrefix, GIF_STAMP_LEN);
-    if (gif89) {
-        strncpy(GifVersionPrefix, GIF89_STAMP, GIF_STAMP_LEN);
-    } else {
-        strncpy(GifVersionPrefix, GIF87_STAMP, GIF_STAMP_LEN);
-    }
     if (EGifPutScreenDesc(GifFileOut,
                           GifFileOut->SWidth,
                           GifFileOut->SHeight,
                           GifFileOut->SColorResolution,
                           GifFileOut->SBackGroundColor,
                           GifFileOut->SColorMap) == GIF_ERROR) {
-        strncpy(GifVersionPrefix, SavedStamp, GIF_STAMP_LEN);
         return (GIF_ERROR);
     }
-    strncpy(GifVersionPrefix, SavedStamp, GIF_STAMP_LEN);
 
     for (i = 0; i < GifFileOut->ImageCount; i++) {
         SavedImage *sp = &GifFileOut->SavedImages[i];
